import base64
from collections.abc import Iterable
import json
import time
from typing import Any, Literal
import uuid

from google import genai
from google.genai import errors as genai_errors
from google.genai import types
from google.genai.types import Content, GenerateContentResponse, GenerateContentResponseUsageMetadata
from openai.types.responses import (
    ResponseFunctionToolCallParam,
    ResponseIncludable,
    ResponseInputImageContentParam,
    ResponseInputItemParam,
    ResponseOutputTextParam,
    ResponseTextConfigParam,
    ResponseUsage,
    ToolParam,
    response_create_params,
)
from openai.types.responses.response_input_item_param import FunctionCallOutput
from openai.types.responses.response_output_message_param import ResponseOutputMessageParam
from openai.types.responses.response_reasoning_item_param import ResponseReasoningItemParam, Summary
from openai.types.responses.response_usage import InputTokensDetails, OutputTokensDetails
from openai.types.shared_params.reasoning import Reasoning

from interop_router.types import ChatMessage, ContextLimitExceededError, RouterResponse, SupportedModelGemini


class GeminiProvider:
    PROVIDER_NAME = "gemini"
    DUMMY_THOUGHT_SIGNATURE = b"skip_thought_signature_validator"

    @staticmethod
    async def create(
        *,
        client: genai.Client,
        input: list[ChatMessage],
        model: SupportedModelGemini,
        include: list[ResponseIncludable] | None = None,
        instructions: str | None = None,
        max_output_tokens: int | None = None,
        parallel_tool_calls: bool | None = None,
        reasoning: Reasoning | None = None,
        temperature: float | None = None,
        text: ResponseTextConfigParam | None = None,
        tool_choice: response_create_params.ToolChoice | None = None,
        tools: Iterable[ToolParam] | None = None,
        truncation: Literal["auto", "disabled"] | None = None,
        background: bool | None = None,
    ) -> RouterResponse:
        preprocessed_input, system_instruction = GeminiProvider._preprocess_input(input)
        gemini_messages = GeminiProvider._convert_input_messages(preprocessed_input)

        # We can have kwargs specific to Gemini in the last message's provider_kwargs
        gemini_kwargs = input[-1].provider_kwargs.get("gemini", {}) if input else {}
        gemini_config, effective_model = GeminiProvider._create_config(
            model=model,
            system_instruction=system_instruction,
            include=include,
            max_output_tokens=max_output_tokens,
            reasoning=reasoning,
            temperature=temperature,
            tools=tools,
            tool_choice=tool_choice,
            gemini_kwargs=gemini_kwargs,
        )

        start_time = time.perf_counter()
        try:
            response = await client.aio.models.generate_content(
                model=effective_model,
                contents=gemini_messages,
                config=gemini_config,
            )
        except genai_errors.ClientError as e:
            if "input_token" in str(e).lower():
                raise ContextLimitExceededError(str(e), provider="gemini", cause=e) from e
            raise
        duration_seconds = time.perf_counter() - start_time

        interop_response = GeminiProvider._convert_response(response)
        interop_response.duration_seconds = duration_seconds
        return interop_response

    @staticmethod
    def _preprocess_input(input: list[ChatMessage]) -> tuple[list[ChatMessage], str]:
        """Removes non-Gemini reasoning messages and
        concats system/developer messages to instructions (and removes them from the input)
        """
        preprocessed_messages = []
        instructions = ""
        for message in input:
            input_message = message.message
            if input_message.get("role", None) in ["system", "developer"]:
                content = input_message.get("content", [])
                if isinstance(content, str):
                    instructions += content + "\n"
                elif isinstance(content, list):
                    for item in content:
                        if isinstance(item, str):
                            instructions += item + "\n"
            elif input_message.get("type", None) == "reasoning" and message.created_by != GeminiProvider.PROVIDER_NAME:
                continue
            else:
                preprocessed_messages.append(message)
        instructions = instructions.strip()
        return preprocessed_messages, instructions

    # region: Input conversion

    @staticmethod
    def _convert_input_messages(input: list[ChatMessage]) -> list[Content]:
        gemini_content: list[Content] = []
        previous_was_function_call = False
        # ids to skip because they were processed already (for image_generation function calls)
        skip_messages: set[str] = set()
        for i in range(len(input)):
            if input[i].id in skip_messages:
                continue
            input_message = input[i].message

            if input_message.get("role") is not None:
                previous_was_function_call = False
                role = input_message.get("role", None)
                content = input_message.get("content", [])
                if isinstance(content, str) and role == "user":
                    gemini_content.append(
                        types.Content(
                            parts=[types.Part(text=content)],
                            role=role,
                        )
                    )
                elif isinstance(content, str) and role == "assistant":
                    thought_signature = GeminiProvider.DUMMY_THOUGHT_SIGNATURE
                    gemini_role = "model" if role == "assistant" else role
                    if i > 0:
                        thought_signature = GeminiProvider._get_thought_signature(input[i - 1].message)
                    gemini_content.append(
                        types.Content(
                            parts=[types.Part(text=content, thought_signature=thought_signature)],
                            role=gemini_role,
                        )
                    )
                elif isinstance(content, list):
                    for content_item in content:
                        if role == "user":
                            if content_item.get("type") == "input_text":
                                text = content_item.get("text", "")
                                gemini_content.append(
                                    types.Content(
                                        parts=[types.Part(text=text)],
                                        role=role,
                                    )
                                )
                            elif content_item.get("type") == "input_image":
                                image_url = content_item.get("image_url")
                                if image_url and image_url.startswith("data:"):
                                    header, base64_data = image_url.split(",", 1)
                                    mime_type = header.split(":")[1].split(";")[0]
                                    image_bytes = base64.b64decode(base64_data)
                                    gemini_content.append(
                                        types.Content(
                                            parts=[types.Part.from_bytes(data=image_bytes, mime_type=mime_type)],
                                            role=role,
                                        )
                                    )
                        elif role == "assistant" and content_item.get("type") == "output_text":
                            text = content_item.get("text", "")
                            thought_signature = GeminiProvider.DUMMY_THOUGHT_SIGNATURE
                            if i > 0:
                                thought_signature = GeminiProvider._get_thought_signature(input[i - 1].message)
                            gemini_content.append(
                                types.Content(
                                    parts=[types.Part(text=text, thought_signature=thought_signature)],
                                    role="model",
                                )
                            )
            elif input_message.get("type") == "function_call":
                if input_message.get("name") == "image_generation":
                    call_id = input_message.get("call_id", "")
                    function_call_output = GeminiProvider._get_corresponding_function_call_output(call_id, input)
                    if function_call_output:
                        skip_messages.add(function_call_output.id)

                        thought_signature = GeminiProvider.DUMMY_THOUGHT_SIGNATURE
                        if i > 0:
                            thought_signature = GeminiProvider._get_thought_signature(input[i - 1].message)

                        # Get the image content from message.output (which is a list)'s image_urls
                        function_call_output_message = function_call_output.message
                        output_content = function_call_output_message.get("output", [])
                        parts: list[types.Part] = []
                        if isinstance(output_content, list):
                            for idx, item in enumerate(output_content):
                                if isinstance(item, dict) and item.get("type") == "input_image":
                                    # Convert each image_url to bytes that Gemini expects
                                    image_url = item.get("image_url", "")
                                    if image_url.startswith("data:"):
                                        header, base64_data = image_url.split(",", 1)
                                        mime_type = header.split(":")[1].split(";")[0]
                                        image_bytes = base64.b64decode(base64_data)
                                        # Only first part gets thought_signature
                                        part_thought_sig = thought_signature if idx == 0 else None
                                        # Create parts with inline data for each image
                                        parts.append(
                                            types.Part(
                                                inline_data=types.Blob(data=image_bytes, mime_type=mime_type),
                                                thought_signature=part_thought_sig,
                                            )
                                        )
                        if parts:
                            gemini_content.append(types.Content(parts=parts, role="model"))
                else:
                    thought_signature = GeminiProvider.DUMMY_THOUGHT_SIGNATURE
                    if not previous_was_function_call and i > 0:
                        thought_signature = GeminiProvider._get_thought_signature(input[i - 1].message)

                    fc_name = input_message.get("name", "")
                    try:
                        fc_arguments = json.loads(input_message.get("arguments", "{}"))
                    except json.JSONDecodeError:
                        fc_arguments = {}
                    function_call = types.FunctionCall(name=fc_name, args=fc_arguments)

                    if not previous_was_function_call:
                        parts = [types.Part(function_call=function_call, thought_signature=thought_signature)]
                    else:
                        parts = [types.Part(function_call=function_call)]

                    gemini_content.append(types.Content(parts=parts, role="model"))
                previous_was_function_call = True
            elif input_message.get("type") == "function_call_output":
                previous_was_function_call = False
                fc_name = GeminiProvider._get_function_call_name_from_id(input, input_message.get("call_id", ""))

                # Gemini expects the function output to be a dict.
                raw_output = input_message.get("output", "{}")
                try:
                    if isinstance(raw_output, str):
                        parsed = json.loads(raw_output)
                        response_dict = parsed if isinstance(parsed, dict) else {"output": parsed}
                    else:
                        response_dict = {"output": raw_output}
                except json.JSONDecodeError:
                    response_dict = {"output": raw_output}

                gemini_content.append(
                    types.Content(
                        parts=[
                            types.Part.from_function_response(
                                name=fc_name,
                                response=response_dict,
                            )
                        ]
                    )
                )

        return gemini_content

    @staticmethod
    def _get_thought_signature(input_message: ResponseInputItemParam) -> bytes:
        """Extract thought_signature from a reasoning message, if it exists."""
        if input_message.get("type") != "reasoning":
            return GeminiProvider.DUMMY_THOUGHT_SIGNATURE

        encrypted_content = input_message.get("encrypted_content")
        thought_signature: bytes = GeminiProvider.DUMMY_THOUGHT_SIGNATURE
        if encrypted_content and isinstance(encrypted_content, str):
            thought_signature = base64.b64decode(encrypted_content)

        return thought_signature

    @staticmethod
    def _get_corresponding_function_call_output(call_id: str, input: list[ChatMessage]) -> ChatMessage | None:
        for message in input:
            input_message = message.message
            if input_message.get("type") == "function_call_output" and input_message.get("call_id") == call_id:
                return message
        return None

    @staticmethod
    def _get_function_call_name_from_id(input: list[ChatMessage], call_id: str) -> str:
        for message in input:
            input_message = message.message
            if input_message.get("type") == "function_call" and input_message.get("call_id") == call_id:
                return input_message.get("name", "")
        return ""

    # endregion

    # region: Gemini config conversion

    @staticmethod
    def _create_config(
        model: SupportedModelGemini,
        system_instruction: str,
        include: list[ResponseIncludable] | None = None,
        max_output_tokens: int | None = None,
        reasoning: Reasoning | None = None,
        temperature: float | None = None,
        tools: Iterable[ToolParam] | None = None,
        tool_choice: response_create_params.ToolChoice | None = None,
        gemini_kwargs: dict[str, Any] | None = None,
    ) -> tuple[types.GenerateContentConfig, str]:
        """Returns the Gemini Config and the model to use based on if image generation is configured."""

        image_config = gemini_kwargs.get("image_config") if gemini_kwargs else None
        gemini_image_config = types.ImageConfig(**image_config) if image_config else None
        image_gen_tool = next(
            (tool for tool in (tools or []) if tool.get("type") == "image_generation"),
            None,
        )
        effective_model = image_gen_tool.get("model") if image_gen_tool else model

        gemini_tools: types.ToolListUnion | None = None
        tool_config: types.ToolConfig | None = None
        if tools:
            gemini_tools = GeminiProvider._convert_tools(tools)
        if tool_choice:
            tool_config = GeminiProvider._convert_tool_choice(tool_choice)

        thinking_config: types.ThinkingConfig | None = None
        include_thoughts: bool | None = None
        thinking_level: types.ThinkingLevel | None = None
        if include and "reasoning.encrypted_content" in include:
            include_thoughts = True

        # Image gen models do not support thinking levels.
        if reasoning and not image_gen_tool:
            match reasoning.get("effort", None):
                case "minimal":
                    thinking_level = types.ThinkingLevel.LOW
                case "low":
                    thinking_level = types.ThinkingLevel.LOW
                case "medium":
                    thinking_level = types.ThinkingLevel.HIGH
                case "high":
                    thinking_level = types.ThinkingLevel.HIGH
                case "xhigh":
                    thinking_level = types.ThinkingLevel.HIGH
                case _:
                    thinking_level = None

        if include_thoughts is not None or thinking_level is not None:
            thinking_config = types.ThinkingConfig(
                include_thoughts=include_thoughts,
                thinking_level=thinking_level,
            )

        config = types.GenerateContentConfig(
            system_instruction=system_instruction or None,
            thinking_config=thinking_config,
            max_output_tokens=max_output_tokens,
            temperature=temperature,
            tools=gemini_tools,
            tool_config=tool_config,
            image_config=gemini_image_config,
            response_modalities=["TEXT", "IMAGE"] if image_gen_tool else None,
            http_options=types.HttpOptions(timeout=3_600_000),
        )
        return config, effective_model

    @staticmethod
    def _convert_tools(tools: Iterable[ToolParam]) -> types.ToolListUnion:
        function_declarations: list[types.FunctionDeclaration] = []
        has_web_search = False
        has_image_generation = False

        for tool in tools:
            tool_type = tool.get("type")
            if tool_type == "image_generation":
                has_image_generation = True
                continue
            if tool_type == "function":
                function_declarations.append(
                    types.FunctionDeclaration(
                        name=tool.get("name"),
                        description=tool.get("description"),
                        parameters_json_schema=tool.get("parameters"),
                    )
                )
            elif tool_type in ("web_search", "web_search_2025_08_26"):
                has_web_search = True

        result: types.ToolListUnion = []
        if function_declarations:
            result.append(types.Tool(function_declarations=function_declarations))
        if has_web_search:
            result.append(types.Tool(google_search=types.GoogleSearch()))
            # Nano banana image gen does not support the UrlContext tool
            if not has_image_generation:
                result.append(types.Tool(url_context=types.UrlContext()))

        return result

    @staticmethod
    def _convert_tool_choice(tool_choice: response_create_params.ToolChoice | None) -> types.ToolConfig | None:
        """Convert OpenAI tool_choice to Gemini ToolConfig."""
        if tool_choice is None:
            return None

        if isinstance(tool_choice, str):
            mode_mapping: dict[str, types.FunctionCallingConfigMode] = {
                "auto": types.FunctionCallingConfigMode.AUTO,
                "none": types.FunctionCallingConfigMode.NONE,
                "required": types.FunctionCallingConfigMode.ANY,
            }
            mode = mode_mapping.get(tool_choice)
            if mode:
                return types.ToolConfig(function_calling_config=types.FunctionCallingConfig(mode=mode))
        elif isinstance(tool_choice, dict) and tool_choice.get("type") == "function":
            function_name = tool_choice.get("name")
            return types.ToolConfig(
                function_calling_config=types.FunctionCallingConfig(
                    mode=types.FunctionCallingConfigMode.ANY,
                    allowed_function_names=[function_name] if function_name else None,
                )
            )

        return None

    # endregion

    # region: Output conversion

    @staticmethod
    def _convert_response(response: GenerateContentResponse) -> RouterResponse:
        output: list[ChatMessage] = []

        if response.candidates:
            for candidate in response.candidates:
                if candidate.content and candidate.content.parts:
                    # Thought summaries will be in a part before where the thought signature is.
                    # It needs to be associated with the ResponseReasoningItemParam. There should only be one per candidate.
                    thought_summary: list[Summary] = []
                    for part in candidate.content.parts:
                        if part.thought and part.text:
                            thought_summary.append(Summary(text=part.text, type="summary_text"))
                            continue

                        if part.thought_signature:
                            encrypted_content = base64.b64encode(part.thought_signature).decode("utf-8")
                            reasoning_item: ResponseReasoningItemParam = {
                                "id": str(uuid.uuid4()),
                                "type": "reasoning",
                                "summary": thought_summary,
                                "status": "completed",
                                "encrypted_content": encrypted_content,
                            }
                            output.append(ChatMessage(message=reasoning_item, created_by=GeminiProvider.PROVIDER_NAME))

                        if part.text and not part.thought:
                            content_item = ResponseOutputTextParam(text=part.text, type="output_text", annotations=[])
                            message_param = ResponseOutputMessageParam(
                                id=str(uuid.uuid4()),
                                content=[content_item],
                                role="assistant",
                                status="completed",
                                type="message",
                            )
                            # Grounding metadata is available on the candidate and only exists if web search was used.
                            grounding_metadata = candidate.grounding_metadata
                            if grounding_metadata:
                                output.append(
                                    ChatMessage(
                                        message=message_param,
                                        created_by=GeminiProvider.PROVIDER_NAME,
                                        interop={
                                            GeminiProvider.PROVIDER_NAME: {
                                                "grounding_metadata": grounding_metadata.model_dump(exclude_none=True)
                                            }
                                        },
                                    )
                                )
                            else:
                                output.append(
                                    ChatMessage(message=message_param, created_by=GeminiProvider.PROVIDER_NAME)
                                )

                        if part.function_call:
                            tool_call = ResponseFunctionToolCallParam(
                                arguments=json.dumps(part.function_call.args or {}),
                                call_id=str(uuid.uuid4()),
                                name=part.function_call.name or "",
                                type="function_call",
                                id=str(uuid.uuid4()),
                                status="completed",
                            )
                            output.append(ChatMessage(message=tool_call, created_by=GeminiProvider.PROVIDER_NAME))

                        # The OpenAI types don't allow inline data (images) in thoughts, so we ignore them.
                        if part.inline_data and part.inline_data.data and not part.thought:
                            call_id = str(uuid.uuid4())
                            mime_type = part.inline_data.mime_type or "image/png"
                            base64_image = base64.b64encode(part.inline_data.data).decode("utf-8")
                            image_url = f"data:{mime_type};base64,{base64_image}"

                            function_call: ResponseFunctionToolCallParam = {
                                "type": "function_call",
                                "call_id": call_id,
                                "name": "image_generation",
                                "arguments": "{}",
                            }
                            function_call_output: FunctionCallOutput = {
                                "type": "function_call_output",
                                "call_id": call_id,
                                "output": [
                                    ResponseInputImageContentParam(
                                        type="input_image", detail="auto", image_url=image_url
                                    )
                                ],
                            }

                            output.append(ChatMessage(message=function_call, created_by=GeminiProvider.PROVIDER_NAME))
                            output.append(
                                ChatMessage(message=function_call_output, created_by=GeminiProvider.PROVIDER_NAME)
                            )

        usage = GeminiProvider._convert_usage(response.usage_metadata) if response.usage_metadata else None
        if output:
            output[-1].original_response = response.model_dump(mode="json")
        return RouterResponse(output=output, usage=usage)

    @staticmethod
    def _convert_usage(usage_metadata: GenerateContentResponseUsageMetadata) -> ResponseUsage:
        return ResponseUsage(
            input_tokens=usage_metadata.prompt_token_count or 0,
            output_tokens=usage_metadata.candidates_token_count or 0,
            total_tokens=usage_metadata.total_token_count or 0,
            input_tokens_details=InputTokensDetails(cached_tokens=-1),
            output_tokens_details=OutputTokensDetails(reasoning_tokens=-1),
        )

    # endregion



from collections.abc import Iterable
from dataclasses import dataclass, field
import datetime
import json
from typing import Any, Literal, Protocol, TypeAlias
import uuid

from openai.types.responses import (
    ResponseError,
    ResponseIncludable,
    ResponseInputItemParam,
    ResponseTextConfigParam,
    ResponseUsage,
    ToolParam,
    response_create_params,
)
from openai.types.responses.response import IncompleteDetails
from openai.types.shared_params.reasoning import Reasoning

ProviderName: TypeAlias = Literal["openai", "gemini", "anthropic"]
CreatedBy: TypeAlias = Literal["user"] | ProviderName

SupportedModelOpenAI: TypeAlias = Literal[
    "gpt-5.2-codex",
    "gpt-5.2",
    "gpt-5.2-2025-12-11",
    "gpt-5.2-chat-latest",
    "gpt-5.2-pro",
    "gpt-5.2-pro-2025-12-11",
    "gpt-5-codex",
    "gpt-5-pro",
    "gpt-5-pro-2025-10-06",
    "gpt-5.1-codex-max",
    "gpt-5.1",
    "gpt-5.1-2025-11-13",
    "gpt-5.1-codex",
    "gpt-5.1-chat-latest",
    "gpt-5",
    "gpt-5-mini",
    "gpt-5-nano",
    "gpt-5-2025-08-07",
    "gpt-5-mini-2025-08-07",
    "gpt-5-nano-2025-08-07",
    "gpt-5-chat-latest",
]

SupportedModelGemini: TypeAlias = Literal[
    "gemini-3-flash-preview",
    "gemini-3-pro-preview",
]

SupportedModelAnthropic: TypeAlias = Literal[
    "claude-opus-4-5-20251101",
    "claude-haiku-4-5-20251001",
    "claude-sonnet-4-5-20250929",
]

SupportedModel: TypeAlias = SupportedModelOpenAI | SupportedModelGemini | SupportedModelAnthropic


@dataclass
class ChatMessage:
    """
    This needs to be a dataclass because of the serialization issues with Pydantic and OpenAI types.
    """

    message: ResponseInputItemParam
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    timestamp: datetime.datetime = field(default_factory=lambda: datetime.datetime.now(datetime.UTC))
    created_by: CreatedBy = "user"
    interop: dict[ProviderName, Any] = field(default_factory=dict)
    metadata: dict[str, Any] = field(default_factory=dict)
    provider_kwargs: dict[ProviderName, dict[str, Any]] = field(default_factory=dict)
    original_response: dict[str, Any] | None = None

    def __post_init__(self) -> None:
        """Validate that timestamp is timezone-aware.

        Raises:
            ValueError: If timestamp does not have timezone information.
        """
        if self.timestamp.tzinfo is None:
            raise ValueError(
                "timestamp must be timezone-aware. Use datetime.now(UTC) or ensure your datetime has timezone info."
            )

    def model_dump_json(self) -> str:
        """Serialize ChatMessage to JSON string.

        Returns:
            JSON string representation of the ChatMessage.
        """
        return json.dumps(
            obj={
                "id": self.id,
                "timestamp": self.timestamp.isoformat(),
                "message": dict(self.message),
                "created_by": self.created_by,
                "interop": self.interop,
                "metadata": self.metadata,
                "provider_kwargs": self.provider_kwargs,
                "original_response": self.original_response,
            }
        )

    @classmethod
    def from_json(cls, json_str: str) -> "ChatMessage":
        """Deserialize ChatMessage from JSON string.

        Args:
            json_str: JSON string containing serialized ChatMessage data.

        Returns:
            ChatMessage instance.

        Raises:
            ValueError: If timestamp string cannot be parsed or is missing timezone.
            json.JSONDecodeError: If json_str is not valid JSON.
            KeyError: If required fields are missing from JSON.
        """
        data = json.loads(json_str)
        timestamp_str = data["timestamp"]
        timestamp = datetime.datetime.fromisoformat(timestamp_str)

        return cls(
            id=data["id"],
            timestamp=timestamp,
            message=data["message"],
            created_by=data.get("created_by", "user"),
            interop=data.get("interop", {}),
            metadata=data.get("metadata", {}),
            provider_kwargs=data.get("provider_kwargs", {}),
            original_response=data.get("original_response"),
        )


@dataclass
class RouterResponse:
    """Response from a router provider call."""

    output: list[ChatMessage]
    error: ResponseError | None = None
    incomplete_details: IncompleteDetails | None = None
    usage: ResponseUsage | None = None
    duration_seconds: float | None = None


class InteropRouterError(Exception):
    def __init__(self, message: str, provider: ProviderName, cause: Exception | None = None) -> None:
        super().__init__(message)
        self.provider = provider
        self.cause = cause


class ContextLimitExceededError(InteropRouterError):
    """Raised when the input context exceeds the provider's limit."""


class ResponsesAPIProtocol(Protocol):
    """
    Protocol defining the interface for provider-agnostic Responses API.
    Uses OpenAI's Responses API types as the common denominator.
    """

    async def create(
        self,
        *,
        client: Any,
        input: list[ChatMessage],
        model: SupportedModel,
        include: list[ResponseIncludable] | None = None,
        instructions: str | None = None,
        max_output_tokens: int | None = None,
        parallel_tool_calls: bool | None = None,
        reasoning: Reasoning | None = None,
        temperature: float | None = None,
        text: ResponseTextConfigParam | None = None,
        tool_choice: response_create_params.ToolChoice | None = None,
        tools: Iterable[ToolParam] | None = None,
        truncation: Literal["auto", "disabled"] | None = None,
        background: bool | None = None,
    ) -> RouterResponse: ...



import asyncio
from collections.abc import Iterable
import json
import time
from typing import Any, Literal, cast
import uuid

import openai
from openai import AsyncOpenAI
from openai._types import omit
from openai.types.responses import (
    Response,
    ResponseCompletedEvent,
    ResponseFunctionToolCallParam,
    ResponseFunctionWebSearch,
    ResponseIncludable,
    ResponseInputImageContentParam,
    ResponseInputItemParam,
    ResponseOutputItem,
    ResponseTextConfigParam,
    response_create_params,
)
from openai.types.responses.response_function_web_search import ActionFind, ActionOpenPage, ActionSearch
from openai.types.responses.response_input_item_param import FunctionCallOutput
from openai.types.responses.tool_param import ToolParam
from openai.types.shared_params.reasoning import Reasoning

from interop_router.types import (
    ChatMessage,
    ContextLimitExceededError,
    ProviderName,
    RouterResponse,
    SupportedModelOpenAI,
)


class OpenAIProvider:
    PROVIDER_NAME = "openai"

    @staticmethod
    async def create(
        *,
        client: AsyncOpenAI,
        input: list[ChatMessage],
        model: SupportedModelOpenAI,
        include: list[ResponseIncludable] | None = None,
        instructions: str | None = None,
        max_output_tokens: int | None = None,
        parallel_tool_calls: bool | None = None,
        reasoning: Reasoning | None = None,
        temperature: float | None = None,
        text: ResponseTextConfigParam | None = None,
        tool_choice: response_create_params.ToolChoice | None = None,
        tools: Iterable[ToolParam] | None = None,
        truncation: Literal["auto", "disabled"] | None = None,
        background: bool | None = None,
    ) -> RouterResponse:
        input_messages = OpenAIProvider._prepare_input_messages(input)
        start_time = time.perf_counter()
        try:
            if background:
                response = await client.responses.create(
                    model=model,
                    input=input_messages,
                    include=include if include is not None else omit,
                    instructions=instructions if instructions is not None else omit,
                    max_output_tokens=max_output_tokens if max_output_tokens is not None else omit,
                    parallel_tool_calls=parallel_tool_calls if parallel_tool_calls is not None else omit,
                    reasoning=reasoning if reasoning is not None else omit,
                    temperature=temperature if temperature is not None else omit,
                    text=text if text is not None else omit,
                    tool_choice=tool_choice if tool_choice is not None else omit,
                    tools=tools if tools is not None else omit,
                    truncation=truncation if truncation is not None else omit,
                    store=True,
                    stream=False,
                    background=True,
                )
                while response.status in {"queued", "in_progress"}:
                    await asyncio.sleep(2)
                    response = await client.responses.retrieve(response.id)
            else:
                response_stream = await client.responses.create(
                    model=model,
                    input=input_messages,
                    include=include if include is not None else omit,
                    instructions=instructions if instructions is not None else omit,
                    max_output_tokens=max_output_tokens if max_output_tokens is not None else omit,
                    parallel_tool_calls=parallel_tool_calls if parallel_tool_calls is not None else omit,
                    reasoning=reasoning if reasoning is not None else omit,
                    temperature=temperature if temperature is not None else omit,
                    text=text if text is not None else omit,
                    tool_choice=tool_choice if tool_choice is not None else omit,
                    tools=tools if tools is not None else omit,
                    truncation=truncation if truncation is not None else omit,
                    store=False,
                    stream=True,
                )
                response: Response | None = None
                async for event in response_stream:
                    if isinstance(event, ResponseCompletedEvent):
                        response = event.response
                        break
                if response is None:
                    raise RuntimeError("Response stream ended without a completed response event.")
        except openai.BadRequestError as e:
            if e.code == "context_length_exceeded":
                raise ContextLimitExceededError(str(e), provider="openai", cause=e) from e
            raise
        duration_seconds = time.perf_counter() - start_time

        converted_output = OpenAIProvider._convert_to_chat_messages(response.output)
        if converted_output:
            converted_output[-1].original_response = response.model_dump(mode="json")
        return RouterResponse(
            output=converted_output,
            error=response.error,
            incomplete_details=response.incomplete_details,
            usage=response.usage,
            duration_seconds=duration_seconds,
        )

    @staticmethod
    def _prepare_input_messages(messages: list[ChatMessage]) -> list[ResponseInputItemParam]:
        """Filters out reasoning from other providers, expands Gemini web search
        metadata into function call pairs, and removes id/status fields.
        """
        expanded: list[ResponseInputItemParam] = []
        for msg in messages:
            is_foreign_reasoning = (
                msg.message.get("type") == "reasoning"
                and msg.created_by != "user"
                and msg.created_by != OpenAIProvider.PROVIDER_NAME
            )
            if is_foreign_reasoning:
                continue
            expanded.extend(OpenAIProvider._expand_gemini_web_search(msg))

        sanitized = []
        for item in expanded:
            sanitized.append(cast(ResponseInputItemParam, {k: v for k, v in item.items() if k not in ("id", "status")}))
        return sanitized

    @staticmethod
    def _expand_gemini_web_search(chat_message: ChatMessage) -> list[ResponseInputItemParam]:
        """Expand a Gemini message with web search metadata into OpenAI function call format.

        If the message has Gemini grounding_metadata, returns:
        - A function_call for "web_search" with args from web_search_queries
        - A function_call_output with grounding_chunks as the result
        - The original message

        Otherwise returns just the original message.
        """
        grounding_metadata: dict[str, Any] | None = chat_message.interop.get("gemini", {}).get("grounding_metadata")
        if grounding_metadata is None:
            return [chat_message.message]

        call_id = str(uuid.uuid4())

        web_search_queries = grounding_metadata.get("web_search_queries", [])
        arguments = json.dumps({"queries": web_search_queries}) if web_search_queries else "{}"

        function_call: ResponseFunctionToolCallParam = {
            "type": "function_call",
            "call_id": call_id,
            "name": "web_search",
            "arguments": arguments,
        }

        grounding_chunks = grounding_metadata.get("grounding_chunks", [])
        function_call_output: FunctionCallOutput = {
            "type": "function_call_output",
            "call_id": call_id,
            "output": json.dumps(grounding_chunks),
        }

        return [function_call, function_call_output, chat_message.message]

    @staticmethod
    def _convert_to_chat_messages(output_items: list[ResponseOutputItem]) -> list[ChatMessage]:
        chat_messages: list[ChatMessage] = []
        for output_item in output_items:
            match output_item.type:
                case "reasoning" | "function_call" | "message":
                    input_param = cast(
                        ResponseInputItemParam, output_item.model_dump(exclude={"status", "id"}, exclude_unset=True)
                    )
                    chat_messages.append(ChatMessage(message=input_param, created_by="openai"))
                case "image_generation_call":
                    chat_messages.extend(OpenAIProvider._convert_image_generation_call(output_item))
                case "web_search_call":
                    output_item = cast(ResponseFunctionWebSearch, output_item)
                    chat_messages.extend(OpenAIProvider._convert_web_search_call(output_item))
                case _:
                    raise ValueError(f"Currently unsupported output item type: {output_item.type}")
        return chat_messages

    @staticmethod
    def _convert_image_generation_call(output_item: ResponseOutputItem) -> list[ChatMessage]:
        """Converts an ImageGenerationCall output item into ResponseFunctionToolCallParam and FunctionCallOutput messages.
        This needs special handling because when store=False, OpenAI does not let you pass the ImageGenerationCall back
        """
        call_id = output_item.id or str(uuid.uuid4())
        function_call: ResponseFunctionToolCallParam = {
            "type": "function_call",
            "call_id": call_id,
            "name": "image_generation",
            "arguments": "{}",
        }
        base64_image = getattr(output_item, "result", None)
        image_url = f"data:image/png;base64,{base64_image}" if base64_image else None
        function_call_output: FunctionCallOutput = {
            "type": "function_call_output",
            "call_id": call_id,
            "output": [ResponseInputImageContentParam(type="input_image", detail="auto", image_url=image_url)],
        }

        extra_fields = output_item.model_dump(
            exclude={"id", "type", "result", "status"},
            exclude_unset=True,
        )
        interop: dict[ProviderName, Any] = {"openai": {"image_generation": extra_fields}} if extra_fields else {}

        return [
            ChatMessage(message=function_call, created_by="openai", interop=interop),
            ChatMessage(message=function_call_output, created_by="openai"),
        ]

    @staticmethod
    def _convert_web_search_call(output_item: ResponseFunctionWebSearch) -> list[ChatMessage]:
        """Converts a built-in web search call into a ResponseFunctionToolCallParam and FunctionCallOutput.
        This is because the built-in web search call data is not persisted, so the model cannot see what it
        searched for or what results were found.
        """
        call_id = output_item.id
        action = output_item.action
        base_message = """The complete result of this tool call was removed. Please call a web search tool if you need the data again. \
Importantly, the web_search tool definition may have changed, including the name of the tool itself. Be sure to use the latest definition."""

        if isinstance(action, ActionSearch):
            arguments = {"type": "search", "query": action.query}
            if action.sources:
                sources_list = [source.url for source in action.sources]
                output_content = base_message + "\n\nSources found:\n" + "\n".join(f"- {url}" for url in sources_list)
            else:
                output_content = base_message
        elif isinstance(action, ActionOpenPage):
            arguments = {"type": "open_page", "url": action.url}
            output_content = base_message
        elif isinstance(action, ActionFind):
            arguments = {"type": "find", "pattern": action.pattern, "url": action.url}
            output_content = base_message
        else:
            arguments = {}
            output_content = base_message

        function_call: ResponseFunctionToolCallParam = {
            "type": "function_call",
            "call_id": call_id,
            "name": "web_search",
            "arguments": json.dumps(arguments),
        }
        function_call_output: FunctionCallOutput = {
            "type": "function_call_output",
            "call_id": call_id,
            "output": output_content,
        }
        return [
            ChatMessage(message=function_call, created_by="openai"),
            ChatMessage(message=function_call_output, created_by="openai"),
        ]



from collections.abc import Iterable
import json
import time
from typing import Any, Literal
import uuid

import anthropic
from anthropic import NOT_GIVEN, AsyncAnthropic
from anthropic.types import (
    Base64ImageSourceParam,
    ImageBlockParam,
    Message,
    MessageParam,
    RedactedThinkingBlock,
    RedactedThinkingBlockParam,
    ServerToolUseBlock,
    ServerToolUseBlockParam,
    TextBlock,
    TextBlockParam,
    ThinkingBlock,
    ThinkingBlockParam,
    ThinkingConfigDisabledParam,
    ThinkingConfigEnabledParam,
    ToolChoiceAnyParam,
    ToolChoiceAutoParam,
    ToolChoiceNoneParam,
    ToolChoiceParam,
    ToolChoiceToolParam,
    ToolResultBlockParam,
    ToolUnionParam,
    ToolUseBlock,
    ToolUseBlockParam,
    URLImageSourceParam,
    Usage,
    WebSearchTool20250305Param,
    WebSearchToolResultBlock,
    WebSearchToolResultBlockParam,
    WebSearchToolResultError,
)
from anthropic.types import ToolParam as AnthropicToolParam
from anthropic.types.beta import BetaWebFetchTool20250910Param, BetaWebFetchToolResultBlockParam
from anthropic.types.web_search_tool_20250305_param import UserLocation as AnthropicUserLocation
from openai.types.responses import (
    ResponseFunctionToolCallParam,
    ResponseIncludable,
    ResponseOutputTextParam,
    ResponseTextConfigParam,
    ResponseUsage,
    ToolParam,
    response_create_params,
)
from openai.types.responses.response_input_item_param import FunctionCallOutput
from openai.types.responses.response_output_message_param import ResponseOutputMessageParam
from openai.types.responses.response_reasoning_item_param import ResponseReasoningItemParam, Summary
from openai.types.responses.response_usage import InputTokensDetails, OutputTokensDetails
from openai.types.shared_params.reasoning import Reasoning

from interop_router.types import (
    ChatMessage,
    ContextLimitExceededError,
    ProviderName,
    RouterResponse,
    SupportedModelAnthropic,
)


class AnthropicProvider:
    PROVIDER_NAME = "anthropic"
    REDACTED_THINKING_BLOCK_ID = "REDACTEDTHINKINGBLOCK"

    @staticmethod
    async def create(
        *,
        client: AsyncAnthropic,
        input: list[ChatMessage],
        model: SupportedModelAnthropic,
        include: list[ResponseIncludable] | None = None,
        instructions: str | None = None,
        max_output_tokens: int | None = None,
        parallel_tool_calls: bool | None = None,
        reasoning: Reasoning | None = None,
        temperature: float | None = None,
        text: ResponseTextConfigParam | None = None,
        tool_choice: response_create_params.ToolChoice | None = None,
        tools: Iterable[ToolParam] | None = None,
        truncation: Literal["auto", "disabled"] | None = None,
        background: bool | None = None,
    ) -> RouterResponse:
        preprocessed_input, system_instruction = AnthropicProvider._preprocess_input(input)
        anthropic_messages = AnthropicProvider._convert_input_messages(preprocessed_input)
        config, extra_headers = AnthropicProvider._create_config(
            max_output_tokens, temperature, reasoning, tool_choice, tools, system_instruction
        )
        start_time = time.perf_counter()
        try:
            async with client.messages.stream(
                messages=anthropic_messages,
                model=model,
                extra_headers=extra_headers if extra_headers else None,
                **config,
            ) as stream:
                async for _text in stream.text_stream:
                    pass
            response = await stream.get_final_message()
        except anthropic.BadRequestError as e:
            if "prompt is too long" in str(e):
                raise ContextLimitExceededError(str(e), provider="anthropic", cause=e) from e
            raise
        duration_seconds = time.perf_counter() - start_time

        interop_response = AnthropicProvider._convert_response(response)
        interop_response.duration_seconds = duration_seconds
        return interop_response

    @staticmethod
    def _preprocess_input(input: list[ChatMessage]) -> tuple[list[ChatMessage], str]:
        preprocessed_messages = []
        instructions = ""
        for message in input:
            input_message = message.message
            if input_message.get("role", None) in ["system", "developer"]:
                content = input_message.get("content", [])
                if isinstance(content, str):
                    instructions += content + "\n"
                elif isinstance(content, list):
                    for item in content:
                        if isinstance(item, str):
                            instructions += item + "\n"
            elif (
                input_message.get("type", None) == "reasoning" and message.created_by != AnthropicProvider.PROVIDER_NAME
            ):
                continue
            else:
                preprocessed_messages.append(message)
        instructions = instructions.strip()
        return preprocessed_messages, instructions

    @staticmethod
    def _convert_input_messages(input: list[ChatMessage]) -> list[MessageParam]:
        message_params: list[MessageParam] = []
        for message in input:
            input_message = message.message

            if input_message.get("role") is not None:
                role = input_message.get("role")
                content = input_message.get("content", [])
                if isinstance(content, str):
                    if role == "user":
                        message_params.append(MessageParam(role="user", content=content))
                    elif role == "assistant":
                        message_params.append(MessageParam(role="assistant", content=content))
                elif isinstance(content, list):
                    content_items = []
                    for content_item in content:
                        if content_item.get("type") == "input_text":
                            text = content_item.get("text", "")
                            content_items.append(TextBlockParam(type="text", text=text))
                        elif content_item.get("type") == "input_image":
                            image_block = AnthropicProvider._convert_input_image(content_item)
                            if image_block:
                                content_items.append(image_block)
                        elif role == "assistant" and content_item.get("type") == "output_text":
                            text = content_item.get("text", "")
                            content_items.append(TextBlockParam(type="text", text=text))
                    if role == "user":
                        message_params.append(MessageParam(role="user", content=content_items))
                    elif role == "assistant":
                        message_params.append(MessageParam(role="assistant", content=content_items))

            elif input_message.get("type") == "reasoning":
                summaries = input_message.get("summary", [])
                if summaries:
                    summary = summaries[0]  # Claude should only have one summary
                    encrypted_content = input_message.get("encrypted_content", "") or ""
                    if summary == AnthropicProvider.REDACTED_THINKING_BLOCK_ID:
                        message_param = MessageParam(
                            role="assistant",
                            content=[
                                RedactedThinkingBlockParam(
                                    data=encrypted_content,
                                    type="redacted_thinking",
                                )
                            ],
                        )
                        message_params.append(message_param)
                    else:
                        thinking_text = summary.get("text", "")
                        message_param = MessageParam(
                            role="assistant",
                            content=[
                                ThinkingBlockParam(
                                    signature=encrypted_content,
                                    thinking=thinking_text,
                                    type="thinking",
                                )
                            ],
                        )
                        message_params.append(message_param)

            elif input_message.get("type") == "function_call":
                anthropic_interop = message.interop.get(AnthropicProvider.PROVIDER_NAME, {})
                call_id = input_message.get("call_id") or ""
                arguments_str = input_message.get("arguments") or "{}"
                arguments = json.loads(arguments_str)
                name = input_message.get("name") or ""

                if anthropic_interop.get("type") == "server_tool_use":
                    block: ServerToolUseBlockParam = ServerToolUseBlockParam(
                        id=call_id,
                        input=arguments,
                        # NOTE: Ignore this type because ServerToolUseBlockParam should allow for name = 'web_fetch' but it does not currently
                        name=name,  # type: ignore
                        type="server_tool_use",
                    )
                else:
                    block = ToolUseBlockParam(
                        id=call_id,
                        input=arguments,
                        name=name,
                        type="tool_use",
                    )
                message_params.append(MessageParam(role="assistant", content=[block]))

            elif input_message.get("type") == "function_call_output":
                anthropic_interop = message.interop.get(AnthropicProvider.PROVIDER_NAME, {})
                call_id = input_message.get("call_id") or ""

                if anthropic_interop.get("type") == "web_search_tool_result":
                    stored_block = anthropic_interop.get("web_search_tool_result_block_param", {})
                    result_block: WebSearchToolResultBlockParam = WebSearchToolResultBlockParam(
                        content=stored_block.get("content"),
                        tool_use_id=call_id,
                        type="web_search_tool_result",
                    )
                    message_params.append(MessageParam(role="assistant", content=[result_block]))
                elif anthropic_interop.get("type") == "web_fetch_tool_result":
                    stored_block = anthropic_interop.get("web_fetch_tool_result", {})
                    fetch_result_block = BetaWebFetchToolResultBlockParam(
                        content=stored_block.get("content"),
                        tool_use_id=call_id,
                        type="web_fetch_tool_result",
                    )
                    # NOTE: The BetaWebFetchToolResultBlockParam is not yet in the main MessageParam union
                    message_params.append(MessageParam(role="assistant", content=[fetch_result_block]))  # type: ignore
                else:
                    output = str(input_message.get("output") or "")
                    tool_result_block = ToolResultBlockParam(
                        tool_use_id=call_id,
                        content=output,
                        type="tool_result",
                    )
                    message_params.append(MessageParam(role="user", content=[tool_result_block]))

        return message_params

    @staticmethod
    def _convert_input_image(content_item: Any) -> ImageBlockParam | None:
        image_url = content_item.get("image_url", "") or ""
        if not image_url:
            return None

        if image_url.startswith("data:"):
            # Parse data URL format: "data:<media_type>;base64,<base64_data>"
            header, base64_data = image_url.split(",", 1)
            media_type = header.split(":")[1].split(";")[0]
            source = Base64ImageSourceParam(type="base64", media_type=media_type, data=base64_data)
            return ImageBlockParam(type="image", source=source)
        else:
            # Regular URL
            source = URLImageSourceParam(type="url", url=image_url)
            return ImageBlockParam(type="image", source=source)

    @staticmethod
    def _create_config(
        max_output_tokens: int | None = None,
        temperature: float | None = None,
        reasoning: Reasoning | None = None,
        tool_choice: response_create_params.ToolChoice | None = None,
        tools: Iterable[ToolParam] | None = None,
        system_instruction: str = "",
    ) -> tuple[dict[str, Any], dict[str, str] | None]:
        max_output_tokens = max_output_tokens if max_output_tokens is not None else 64000

        thinking_config_param = ThinkingConfigDisabledParam(type="disabled")
        thinking_budget_tokens = 0
        if reasoning is not None:
            reasoning_effort = reasoning.get("effort", "none")
            thinking_budget_tokens = 1024
            match reasoning_effort:
                case "none":
                    thinking_budget_tokens = 1024  # The minimum if thinking is enabled is 1024
                case "low":
                    thinking_budget_tokens = 2_000
                case "medium":
                    thinking_budget_tokens = 8_000
                case "high":
                    thinking_budget_tokens = 16_000
                case "xhigh":
                    # 64k is the maximum, but this does not leave room for output tokens
                    thinking_budget_tokens = 32_000

            thinking_config_param = ThinkingConfigEnabledParam(budget_tokens=thinking_budget_tokens, type="enabled")

        anthropic_tools, has_web_fetch = AnthropicProvider._convert_tools(tools) if tools else ([], False)
        anthropic_tool_choice = AnthropicProvider._convert_tool_choice(tool_choice)

        extra_headers: dict[str, str] | None = None
        if has_web_fetch:
            extra_headers = {"anthropic-beta": "web-fetch-2025-09-10,interleaved-thinking-2025-05-14"}

        config = {
            "max_tokens": max_output_tokens,
            "temperature": temperature if temperature is not None else NOT_GIVEN,
            "system": system_instruction if system_instruction else NOT_GIVEN,
            "thinking": thinking_config_param,
            "tools": anthropic_tools,
            "tool_choice": anthropic_tool_choice if anthropic_tool_choice is not None else NOT_GIVEN,
            "timeout": 3600,
        }
        return config, extra_headers

    @staticmethod
    def _convert_tools(
        tools: Iterable[ToolParam],
    ) -> tuple[list[ToolUnionParam | BetaWebFetchTool20250910Param], bool]:
        anthropic_tools: list[ToolUnionParam | BetaWebFetchTool20250910Param] = []
        web_search_tool: ToolParam | None = None
        has_web_fetch = False

        for tool in tools:
            tool_type = tool.get("type")
            if tool_type == "function":
                name = tool.get("name") or ""
                description = tool.get("description") or ""
                parameters = tool.get("parameters") or {"type": "object", "properties": {}}
                anthropic_tools.append(
                    AnthropicToolParam(
                        name=name,
                        description=description,
                        input_schema=parameters,
                    )
                )
            elif tool_type in ("web_search", "web_search_2025_08_26"):
                web_search_tool = tool

        if web_search_tool:
            web_search_param: WebSearchTool20250305Param = {
                "type": "web_search_20250305",
                "name": "web_search",
            }
            filters = web_search_tool.get("filters")
            if filters and isinstance(filters, dict):
                allowed_domains = filters.get("allowed_domains")
                if allowed_domains:
                    web_search_param["allowed_domains"] = allowed_domains
            user_location = web_search_tool.get("user_location")
            if user_location and isinstance(user_location, dict):
                anthropic_user_location: AnthropicUserLocation = {"type": "approximate"}
                if user_location.get("city"):
                    anthropic_user_location["city"] = user_location.get("city")
                if user_location.get("country"):
                    anthropic_user_location["country"] = user_location.get("country")
                if user_location.get("region"):
                    anthropic_user_location["region"] = user_location.get("region")
                if user_location.get("timezone"):
                    anthropic_user_location["timezone"] = user_location.get("timezone")
                web_search_param["user_location"] = anthropic_user_location
            anthropic_tools.append(web_search_param)

            web_fetch_param: BetaWebFetchTool20250910Param = {
                "type": "web_fetch_20250910",
                "name": "web_fetch",
                "citations": {"enabled": False},
            }
            if filters and isinstance(filters, dict):
                allowed_domains = filters.get("allowed_domains")
                if allowed_domains:
                    web_fetch_param["allowed_domains"] = allowed_domains
            anthropic_tools.append(web_fetch_param)
            has_web_fetch = True

        return anthropic_tools, has_web_fetch

    @staticmethod
    def _convert_tool_choice(
        tool_choice: response_create_params.ToolChoice | None,
    ) -> ToolChoiceParam | None:
        if tool_choice is None:
            return None

        if isinstance(tool_choice, str):
            match tool_choice:
                case "auto":
                    return ToolChoiceAutoParam(type="auto")
                case "none":
                    return ToolChoiceNoneParam(type="none")
                case "required":
                    return ToolChoiceAnyParam(type="any")
        elif isinstance(tool_choice, dict) and tool_choice.get("type") == "function":
            function_name = tool_choice.get("name") or ""
            return ToolChoiceToolParam(type="tool", name=function_name)

        return None

    @staticmethod
    def _convert_response(response: Message) -> RouterResponse:
        output: list[ChatMessage] = []
        usage = AnthropicProvider._convert_usage(response.usage)
        for content_block in response.content:
            if isinstance(content_block, TextBlock):
                if content_block.type == "web_fetch_tool_result":
                    # Currently the web fetch tool result is returned as a text block and doesn't have a dedicated type
                    block_content = content_block.content
                    interop_data: dict[ProviderName, Any] = {
                        AnthropicProvider.PROVIDER_NAME: {
                            "type": "web_fetch_tool_result",
                            "web_fetch_tool_result": content_block.model_dump(),
                        }
                    }
                    generic_output = json.dumps(block_content)
                    function_call_output = FunctionCallOutput(
                        call_id=content_block.tool_use_id,
                        output=generic_output,
                        type="function_call_output",
                    )
                    output.append(
                        ChatMessage(
                            message=function_call_output,
                            created_by=AnthropicProvider.PROVIDER_NAME,
                            interop=interop_data,
                        )
                    )
                else:
                    content_item = ResponseOutputTextParam(text=content_block.text, type="output_text", annotations=[])
                    message_param = ResponseOutputMessageParam(
                        id=str(uuid.uuid4()),
                        content=[content_item],
                        role="assistant",
                        status="completed",
                        type="message",
                    )
                    output.append(ChatMessage(message=message_param, created_by=AnthropicProvider.PROVIDER_NAME))
            elif isinstance(content_block, ThinkingBlock):
                reasoning_item: ResponseReasoningItemParam = {
                    "id": str(uuid.uuid4()),
                    "type": "reasoning",
                    "summary": [Summary(text=content_block.thinking, type="summary_text")],
                    "status": "completed",
                    "encrypted_content": content_block.signature,
                }
                output.append(ChatMessage(message=reasoning_item, created_by=AnthropicProvider.PROVIDER_NAME))
            elif isinstance(content_block, RedactedThinkingBlock):
                reasoning_item: ResponseReasoningItemParam = {
                    "id": str(uuid.uuid4()),
                    "type": "reasoning",
                    "summary": [Summary(text=AnthropicProvider.REDACTED_THINKING_BLOCK_ID, type="summary_text")],
                    "status": "completed",
                    "encrypted_content": content_block.data,
                }
                output.append(ChatMessage(message=reasoning_item, created_by=AnthropicProvider.PROVIDER_NAME))
            elif isinstance(content_block, ToolUseBlock):
                tool_call = ResponseFunctionToolCallParam(
                    call_id=content_block.id,
                    arguments=json.dumps(content_block.input),
                    name=content_block.name,
                    type="function_call",
                    id=str(uuid.uuid4()),
                    status="completed",
                )
                output.append(ChatMessage(message=tool_call, created_by=AnthropicProvider.PROVIDER_NAME))
            elif isinstance(content_block, ServerToolUseBlock):
                tool_call = ResponseFunctionToolCallParam(
                    call_id=content_block.id,
                    arguments=json.dumps(content_block.input),
                    name=content_block.name,
                    type="function_call",
                    id=str(uuid.uuid4()),
                    status="completed",
                )
                interop_data: dict[ProviderName, Any] = {
                    AnthropicProvider.PROVIDER_NAME: {
                        "type": "server_tool_use",
                    }
                }
                output.append(
                    ChatMessage(message=tool_call, created_by=AnthropicProvider.PROVIDER_NAME, interop=interop_data)
                )
            elif isinstance(content_block, WebSearchToolResultBlock):
                generic_output = AnthropicProvider._convert_web_search_tool_result_block(content_block)
                function_call_output = FunctionCallOutput(
                    call_id=content_block.tool_use_id,
                    output=generic_output,
                    type="function_call_output",
                )
                interop_data: dict[ProviderName, Any] = {
                    AnthropicProvider.PROVIDER_NAME: {
                        "type": "web_search_tool_result",
                        "web_search_tool_result_block_param": content_block.model_dump(),
                    }
                }
                output.append(
                    ChatMessage(
                        message=function_call_output,
                        created_by=AnthropicProvider.PROVIDER_NAME,
                        interop=interop_data,
                    )
                )

        if output:
            output[-1].original_response = response.model_dump(mode="json")
        return RouterResponse(output=output, usage=usage)

    @staticmethod
    def _convert_web_search_tool_result_block(
        block: WebSearchToolResultBlock,
    ) -> str:
        if isinstance(block.content, WebSearchToolResultError):
            return ""

        results = [
            {
                "title": result.title,
                "url": result.url,
                "page_age": result.page_age,
            }
            for result in block.content
        ]
        return json.dumps(results)

    @staticmethod
    def _convert_usage(usage: Usage) -> ResponseUsage:
        input_tokens = usage.input_tokens
        output_tokens = usage.output_tokens
        total_tokens = input_tokens + output_tokens
        cached_tokens = usage.cache_read_input_tokens or 0

        input_tokens_details = InputTokensDetails(cached_tokens=cached_tokens)
        output_tokens_details = OutputTokensDetails(reasoning_tokens=0)

        return ResponseUsage(
            input_tokens=input_tokens,
            input_tokens_details=input_tokens_details,
            output_tokens=output_tokens,
            output_tokens_details=output_tokens_details,
            total_tokens=total_tokens,
        )



from collections.abc import Iterable
from typing import Any, Literal, cast, get_args

from anthropic import AsyncAnthropic
from google import genai
from openai import AsyncOpenAI
from openai.types.responses import ResponseIncludable, ResponseTextConfigParam, response_create_params
from openai.types.responses.tool_param import ToolParam
from openai.types.shared_params.reasoning import Reasoning

from interop_router.anthropic_provider import AnthropicProvider
from interop_router.gemini_provider import GeminiProvider
from interop_router.openai_provider import OpenAIProvider
from interop_router.types import (
    ChatMessage,
    ProviderName,
    RouterResponse,
    SupportedModel,
    SupportedModelAnthropic,
    SupportedModelGemini,
    SupportedModelOpenAI,
)


class Router:
    """Router that dispatches API calls to the appropriate provider based on model type."""

    def __init__(self) -> None:
        self._clients: dict[ProviderName, Any] = {}

    def register(self, provider_name: ProviderName, client: AsyncOpenAI | genai.Client | AsyncAnthropic) -> None:
        self._clients[provider_name] = client

    def _get_provider_for_model(self, model: SupportedModel) -> ProviderName:
        if model in get_args(SupportedModelOpenAI):
            return "openai"
        if model in get_args(SupportedModelGemini):
            return "gemini"
        if model in get_args(SupportedModelAnthropic):
            return "anthropic"
        raise ValueError(f"Unknown model: {model}")

    async def create(
        self,
        *,
        input: list[ChatMessage],
        model: SupportedModel,
        include: list[ResponseIncludable] | None = None,
        instructions: str | None = None,
        max_output_tokens: int | None = None,
        parallel_tool_calls: bool | None = None,
        reasoning: Reasoning | None = None,
        temperature: float | None = None,
        text: ResponseTextConfigParam | None = None,
        tool_choice: response_create_params.ToolChoice | None = None,
        tools: Iterable[ToolParam] | None = None,
        truncation: Literal["auto", "disabled"] | None = None,
        background: bool | None = None,
    ) -> RouterResponse:
        """Create a response using the appropriate provider for the given model.

        Args:
            input: List of chat messages.
            model: The model to use for generation.
            include: Optional list of response includables.
            instructions: Optional system instructions.
            max_output_tokens: Optional maximum output tokens.
            parallel_tool_calls: Optional flag for parallel tool calls.
            reasoning: Optional reasoning configuration.
            temperature: Optional temperature setting.
            text: Optional text configuration.
            tool_choice: Optional tool choice configuration.
            tools: Optional list of tools.
            truncation: Optional truncation setting.
            background: Whether to run the model response in the background.

        Returns:
            The response from the provider.

        Raises:
            ValueError: If no client is registered for the required provider.
        """
        if model in get_args(SupportedModelOpenAI):
            client = self._clients.get("openai")
            if client is None:
                raise ValueError("No client registered for provider: openai")
            return await OpenAIProvider.create(
                client=client,
                input=input,
                model=cast(SupportedModelOpenAI, model),
                include=include,
                instructions=instructions,
                max_output_tokens=max_output_tokens,
                parallel_tool_calls=parallel_tool_calls,
                reasoning=reasoning,
                temperature=temperature,
                text=text,
                tool_choice=tool_choice,
                tools=tools,
                truncation=truncation,
                background=background,
            )

        if model in get_args(SupportedModelGemini):
            client = self._clients.get("gemini")
            if client is None:
                raise ValueError("No client registered for provider: gemini")
            return await GeminiProvider.create(
                client=client,
                input=input,
                model=cast(SupportedModelGemini, model),
                include=include,
                instructions=instructions,
                max_output_tokens=max_output_tokens,
                parallel_tool_calls=parallel_tool_calls,
                reasoning=reasoning,
                temperature=temperature,
                text=text,
                tool_choice=tool_choice,
                tools=tools,
                truncation=truncation,
            )

        if model in get_args(SupportedModelAnthropic):
            client = self._clients.get("anthropic")
            if client is None:
                raise ValueError("No client registered for provider: anthropic")
            return await AnthropicProvider.create(
                client=client,
                input=input,
                model=cast(SupportedModelAnthropic, model),
                include=include,
                instructions=instructions,
                max_output_tokens=max_output_tokens,
                parallel_tool_calls=parallel_tool_calls,
                reasoning=reasoning,
                temperature=temperature,
                text=text,
                tool_choice=tool_choice,
                tools=tools,
                truncation=truncation,
            )

        raise ValueError(f"Unknown model: {model}")
